from typing import Optional
import torch
from GANDLF.grad_clipping.clip_gradients import dispatch_clip_grad_


class GradScaler:
    def __init__(self):
        """Initializes a GradScaler object with a PyTorch GradScaler."""
        self._scaler = torch.cuda.amp.GradScaler()

    def __call__(
        self,
        loss: torch.Tensor,
        optimizer: torch.optim.Optimizer,
        clip_grad: Optional[float] = None,
        clip_mode: Optional[str] = "norm",
        parameters: Optional[torch.Tensor] = None,
        create_graph: Optional[bool] = False,
    ):
        """
        Scales the loss and performs backward pass through the computation graph.

        Args:
            loss (torch.Tensor): The loss tensor to scale and backpropagate.
            optimizer (torch.optim.Optimizer): The optimizer to step after backpropagation.
            clip_grad (Optional[float], optional): The clipping value/factor/norm, mode dependent. Defaults to None.
            clip_mode (Optional[str], optional): The mode of clipping. Defaults to "norm".
            parameters (Optional[torch.Tensor], optional): The model parameters to be clipped. Defaults to None.
            create_graph (Optional[bool], optional): Whether to create the graph. Defaults to False.
        """
        self._scaler.scale(loss).backward(create_graph=create_graph)
        if clip_grad is not None:
            assert parameters is not None
            # unscale the gradients of optimizer's assigned params in-place
            self._scaler.unscale_(optimizer)
            if (clip_mode is None) or (str(clip_mode).lower() == "none"):
                clip_mode = "norm"  # default, in case none gets passed
            dispatch_clip_grad_(parameters, clip_grad, mode=clip_mode)
        self._scaler.step(optimizer)
        self._scaler.update()

    def state_dict(self):
        """
        Returns the state dict of the underlying GradScaler.
        """
        return self._scaler.state_dict()

    def load_state_dict(self, state_dict):
        """
        Loads the state dict of the underlying GradScaler.

        Args:
            state_dict (dict): The state dict to load.
        """
        self._scaler.load_state_dict(state_dict)


def model_parameters_exclude_head(
    model: torch.nn.Module, clip_mode: Optional[str] = None
):
    """
    Returns the parameters of a PyTorch model excluding the last two layers (the head).

    Args:
        model (torch.nn.Module): The PyTorch model to get the parameters from.
        clip_mode (Optional[str], optional): The mode of clipping. Defaults to None.

    Returns:
        Iterable: The model parameters excluding the last two layers if clip_mode is 'agc', otherwise all parameters.
    """
    exclude_head = False
    clip_mode = str(clip_mode).lower() if clip_mode is not None else None
    if clip_mode == "agc":
        exclude_head = True
    if exclude_head:
        return [p for p in model.parameters()][:-2]
    else:
        return model.parameters()



"""
All the metrics are to be called from here
"""
from typing import Union

import torch
from sklearn.metrics import balanced_accuracy_score
import numpy as np
import torchmetrics as tm
from ..utils import get_output_from_calculator


def classification_accuracy(
    prediction: torch.Tensor, target: torch.Tensor, params: dict
) -> torch.Tensor:
    """
    This function computes the classification accuracy.

    Args:
        prediction (torch.Tensor): The prediction of the model.
        target (torch.Tensor): The ground truth labels.
        params (dict): The parameter dictionary containing training and data information.

    Returns:
        torch.Tensor: The classification accuracy.
    """
    predicted_classes = prediction
    if params["problem_type"] == "classification":
        predicted_classes = torch.argmax(prediction, 1)

    acc = torch.sum(predicted_classes == target.squeeze()) / len(target)
    return acc


def balanced_acc_score(
    prediction: torch.Tensor, target: torch.Tensor, params: dict
) -> torch.Tensor:
    """
    This function computes the balanced accuracy.

    Args:
        prediction (torch.Tensor): The prediction of the model.
        target (torch.Tensor): The ground truth labels.
        params (dict): The parameter dictionary containing training and data information.

    Returns:
        torch.Tensor: The balanced accuracy.
    """
    predicted_classes = prediction
    if params["problem_type"] == "classification":
        predicted_classes = torch.argmax(prediction, 1)

    return torch.from_numpy(
        np.array(balanced_accuracy_score(predicted_classes.cpu(), target.cpu()))
    )


def per_label_accuracy(
    prediction: torch.Tensor, target: torch.Tensor, params: dict
) -> torch.Tensor:
    """
    This function computes the per class accuracy.

    Args:
        prediction (torch.Tensor): The prediction of the model.
        target (torch.Tensor): The ground truth labels.
        params (dict): The parameter dictionary containing training and data information.

    Returns:
        torch.Tensor: The per class accuracy.
    """
    if params["problem_type"] == "classification":
        # ensure this works for multiple batches
        output_accuracy = torch.zeros(len(params["model"]["class_list"]))
        for output_batch, label_batch in zip(prediction, target):
            predicted_classes = torch.Tensor([0] * len(params["model"]["class_list"]))
            label_cpu = torch.Tensor([0] * len(params["model"]["class_list"]))
            predicted_classes[torch.argmax(output_batch, 0).cpu().item()] = 1
            label_cpu[label_batch.cpu().item()] = 1
            output_accuracy += (predicted_classes == label_cpu).type(torch.float)
        return output_accuracy / len(prediction)
    else:
        return balanced_acc_score(prediction, target, params)


def overall_stats(
    prediction: torch.Tensor, target: torch.Tensor, params: dict
) -> dict[str, Union[float, list]]:
    """
    Generates a dictionary of metrics calculated on the overall predictions and ground truths.

    Args:
        predictions (torch.Tensor): The prediction of the model.
        target (torch.Tensor): The ground truth labels.
        params (dict): The parameter dictionary containing training and data information.

    Returns:
        dict: A dictionary of metrics.
    """
    prediction = prediction.type(torch.float)
    target = target.type(torch.float) * params["scaling_factor"]
    assert (
        params["problem_type"] == "regression"
    ), "Only regression is supported for these stats"

    output_metrics = {}

    reduction_types_keys = {"mean": "mean", "sum": "sum", "none": "none"}
    # metrics that need the "reduction" parameter
    for reduction_type, reduction_type_key in reduction_types_keys.items():
        calculators = {
            "cosinesimilarity": tm.CosineSimilarity(reduction=reduction_type_key)
        }
        for metric_name, calculator in calculators.items():
            output_metrics[
                f"{metric_name}_{reduction_type}"
            ] = get_output_from_calculator(prediction, target, calculator)
    # metrics that do not have any "reduction" parameter
    calculators = {
        "mse": tm.MeanSquaredError(),
        "mae": tm.MeanAbsoluteError(),
        "pearson": tm.PearsonCorrCoef(),
        "spearman": tm.SpearmanCorrCoef(),
    }
    for metric_name, calculator in calculators.items():
        output_metrics[metric_name] = get_output_from_calculator(
            prediction, target, calculator
        )

    return output_metrics
